@@ -23,14 +23,15 @@
 
 #include <config.h>
 
-#include "panel-compatibility.h"
+#include "string.h"
 
-#include <libgnome/gnome-i18n.h>
+#include "panel-compatibility.h"
 
 #include "panel-profile.h"
 #include "panel-menu-bar.h"
 #include "panel-applet-frame.h"
 #include "panel-globals.h"
+#include "panel-util.h"
 
 typedef enum {
 	PANEL_ORIENT_UP    = GNOME_Vertigo_PANEL_ORIENT_UP,
@@ -498,12 +499,9 @@ panel_compatibility_migrate_menu_panel_settings (GConfClient *client,
 						 const char  *panel_dir)
 {
 	const char *key;
-	const char *profile;
 	const char *toplevel_id;
 	char       *id;
 
-	profile = panel_profile_get_name ();
-
 	key = panel_gconf_sprintf ("%s/expand", toplevel_dir);
 	gconf_client_set_bool (client, key, TRUE, NULL);
 
@@ -522,7 +520,7 @@ panel_compatibility_migrate_menu_panel_settings (GConfClient *client,
 	/* window menu on far left corner */
         id = panel_profile_prepare_object_with_id (PANEL_OBJECT_BONOBO, toplevel_id, 0, TRUE);
 
-	key = panel_gconf_full_key (PANEL_GCONF_APPLETS, profile, id, "bonobo_iid");
+	key = panel_gconf_full_key (PANEL_GCONF_APPLETS, id, "bonobo_iid");
         gconf_client_set_string (client, key, "OAFIID:GNOME_WindowMenuApplet", NULL);
 
         panel_profile_add_to_list (PANEL_GCONF_APPLETS, id);
@@ -581,7 +579,6 @@ panel_compatibility_migrate_panel_settings (GConfClient *client,
 					    const char  *panel_id,
 					    gboolean    *is_drawer)
 {
-	const char *profile;
 	const char *key;
 	char       *toplevel_id;
 	char       *toplevel_dir;
@@ -594,12 +591,10 @@ panel_compatibility_migrate_panel_settings (GConfClient *client,
 	gboolean    enable_arrows;
 	gboolean    auto_hide;
 
-	profile = panel_profile_get_name ();
-
 	toplevel_id = panel_profile_find_new_id (PANEL_GCONF_TOPLEVELS);
 
-	toplevel_dir = g_strdup_printf (PANEL_CONFIG_DIR "/%s/toplevels/%s", profile, toplevel_id);
-	panel_dir    = g_strdup_printf (PANEL_CONFIG_DIR "/%s/panels/%s", profile, panel_id);
+	toplevel_dir = g_strdup_printf (PANEL_CONFIG_DIR "/toplevels/%s", toplevel_id);
+	panel_dir    = g_strdup_printf (PANEL_CONFIG_DIR "/panels/%s", panel_id);
 
 	panel_gconf_associate_schemas_in_dir (
 			client, toplevel_dir, PANEL_SCHEMAS_DIR "/toplevels");
@@ -668,31 +663,22 @@ panel_compatibility_migrate_panel_id (GConfClient       *client,
 				      const char        *object_id,
 				      GHashTable        *panel_id_hash)
 {
-	const char *profile;
 	const char *key;
 	char       *panel_id;
 	char       *toplevel_id;
 	gboolean    retval = FALSE;
 
-	profile = panel_profile_get_name ();
-
 	/* panel_id -> toplevel_id */
-	key = panel_gconf_full_key (key_type, profile, object_id, "toplevel_id");
-	toplevel_id = gconf_client_get_string (client, key, NULL);
-
-	key = panel_gconf_full_key (key_type, profile, object_id, "panel_id");
+	key = panel_gconf_full_key (key_type, object_id, "panel_id");
 	panel_id = gconf_client_get_string (client, key, NULL);
 
-	if (!toplevel_id && panel_id &&
-	    (toplevel_id = g_hash_table_lookup (panel_id_hash, panel_id))) {
-		key = panel_gconf_full_key (key_type, profile, object_id, "toplevel_id");
+	if (panel_id && (toplevel_id = g_hash_table_lookup (panel_id_hash, panel_id))) {
+		key = panel_gconf_full_key (key_type, object_id, "toplevel_id");
 		gconf_client_set_string (client, key, toplevel_id, NULL);
 
-		toplevel_id = NULL;
 		retval = TRUE;
 	}
 
-	g_free (toplevel_id);
 	g_free (panel_id);
 
 	return retval;
@@ -704,42 +690,39 @@ panel_compatibility_migrate_drawer_settings (GConfClient       *client,
 					     const char        *object_id,
 					     GHashTable        *panel_id_hash)
 {
-	const char *profile;
 	const char *key;
 	char       *toplevel_id;
 	char       *panel_id;
 	char       *custom_icon;
 	char       *pixmap;
 
-	profile = panel_profile_get_name ();
-
 	/* unique-drawer-panel-id -> attached_toplevel_id */
-	key = panel_gconf_full_key (key_type, profile, object_id, "attached_toplevel_id");
+	key = panel_gconf_full_key (key_type, object_id, "attached_toplevel_id");
 	toplevel_id = gconf_client_get_string (client, key, NULL);
 
-	key = panel_gconf_full_key (key_type, profile, object_id, "unique-drawer-panel-id");
+	key = panel_gconf_full_key (key_type, object_id, "unique-drawer-panel-id");
 	panel_id = gconf_client_get_string (client, key, NULL);
 
 	if (!toplevel_id && panel_id &&
 	    (toplevel_id = g_hash_table_lookup (panel_id_hash, panel_id))) {
-		key = panel_gconf_full_key (key_type, profile, object_id, "attached_toplevel_id");
+		key = panel_gconf_full_key (key_type, object_id, "attached_toplevel_id");
 		gconf_client_set_string (client, key, toplevel_id, NULL);
 
 		toplevel_id = NULL;
 	}
 
 	/* pixmap -> custom_icon */	
-	key = panel_gconf_full_key (key_type, profile, object_id, "custom_icon");
+	key = panel_gconf_full_key (key_type, object_id, "custom_icon");
 	custom_icon = gconf_client_get_string (client, key, NULL);
 
-	key = panel_gconf_full_key (key_type, profile, object_id, "pixmap");
+	key = panel_gconf_full_key (key_type, object_id, "pixmap");
 	pixmap = gconf_client_get_string (client, key, NULL);
 
 	if (!custom_icon && pixmap) {
-		key = panel_gconf_full_key (key_type, profile, object_id, "custom_icon");
+		key = panel_gconf_full_key (key_type, object_id, "custom_icon");
 		gconf_client_set_string (client, key, pixmap, NULL);
 
-		key = panel_gconf_full_key (key_type, profile, object_id, "use_custom_icon");
+		key = panel_gconf_full_key (key_type, object_id, "use_custom_icon");
 		gconf_client_set_bool (client, key, TRUE, NULL);
 	}
 
@@ -754,44 +737,41 @@ panel_compatibility_migrate_menu_button_settings (GConfClient       *client,
 						  PanelGConfKeyType  key_type,
 						  const char        *object_id)
 {
-	const char *profile;
 	const char *key;
 	gboolean    use_custom_icon;
 	gboolean    use_menu_path;
 	char       *custom_icon;
 	char       *menu_path;
 
-	profile = panel_profile_get_name ();
-
 	/* custom-icon -> use_custom_icon */
-	key = panel_gconf_full_key (key_type, profile, object_id, "custom-icon");
+	key = panel_gconf_full_key (key_type, object_id, "custom-icon");
 	use_custom_icon = gconf_client_get_bool (client, key, NULL);
 
-	key = panel_gconf_full_key (key_type, profile, object_id, "use_custom_icon");
+	key = panel_gconf_full_key (key_type, object_id, "use_custom_icon");
 	gconf_client_set_bool (client, key, use_custom_icon, NULL);
 
 	/* custom-icon-file -> custom_icon */
-	key = panel_gconf_full_key (key_type, profile, object_id, "custom-icon-file");
+	key = panel_gconf_full_key (key_type, object_id, "custom-icon-file");
 	custom_icon = gconf_client_get_string (client, key, NULL);
 
 	if (custom_icon) {
-		key = panel_gconf_full_key (key_type, profile, object_id, "custom_icon");
+		key = panel_gconf_full_key (key_type, object_id, "custom_icon");
 		gconf_client_set_string (client, key, custom_icon, NULL);
 	}
 
 	/* main_menu -> ! use_menu_path */
-	key = panel_gconf_full_key (key_type, profile, object_id, "main-menu");
+	key = panel_gconf_full_key (key_type, object_id, "main-menu");
 	use_menu_path = ! gconf_client_get_bool (client, key, NULL);
 
-	key = panel_gconf_full_key (key_type, profile, object_id, "use_menu_path");
+	key = panel_gconf_full_key (key_type, object_id, "use_menu_path");
 	gconf_client_set_bool (client, key, use_menu_path, NULL);
 
 	/* path -> menu_path */
-	key = panel_gconf_full_key (key_type, profile, object_id, "path");
+	key = panel_gconf_full_key (key_type, object_id, "path");
 	menu_path = gconf_client_get_string (client, key, NULL);
 
 	if (menu_path) {
-		key = panel_gconf_full_key (key_type, profile, object_id, "menu_path");
+		key = panel_gconf_full_key (key_type, object_id, "menu_path");
 		gconf_client_set_string (client, key, menu_path, NULL);
 	}
 
@@ -805,12 +785,9 @@ panel_compatibility_migrate_objects (GConfClient       *client,
 				     GHashTable        *panel_id_hash)
 {
 	const char *key;
-	const char *profile;
 	GSList     *l, *objects;
 
-	profile = panel_profile_get_name ();
-
-	key = panel_gconf_general_key (profile, panel_gconf_key_type_to_id_list (key_type));
+	key = panel_gconf_general_key (panel_gconf_key_type_to_id_list (key_type));
 	objects = gconf_client_get_list (client, key, GCONF_VALUE_STRING, NULL);
 
 	for (l = objects; l; l = l->next) {
@@ -823,7 +800,7 @@ panel_compatibility_migrate_objects (GConfClient       *client,
 			continue;
 		}
 
-		key = panel_gconf_full_key (key_type, profile, id, "object_type");
+		key = panel_gconf_full_key (key_type, id, "object_type");
 		object_type_str = gconf_client_get_string (client, key, NULL);
 
 		if (panel_profile_map_object_type_string (object_type_str, &object_type)) {
@@ -846,6 +823,79 @@ panel_compatibility_migrate_objects (GConfClient       *client,
 	g_slist_free (objects);
 }
 
+/* Major hack, but we now set toplevel_id_list in the defaults database,
+ * so we need to figure out if its actually set in the users database.
+ */
+
+static GConfEngine *
+get_homedir_source (void)
+{
+	GConfEngine *engine;
+	GError      *error = NULL;
+	char        *source;
+
+	source = g_strdup_printf ("xml:readwrite:%s/.gconf", g_get_home_dir ());
+
+	if (!(engine = gconf_engine_get_for_address (source, &error))) {
+#if 0
+		g_warning ("Cannot get GConf source '%s': %s\n",
+			   source, error->message);
+#endif
+		g_error_free (error);
+		g_free (source);
+		return NULL;
+	}
+
+	g_free (source);
+
+	return engine;
+}
+
+static gboolean
+is_general_key_set (GConfEngine *engine,
+		    const char  *config_dir,
+		    const char  *general_key)
+{
+	GConfEntry *entry;
+	const char *key;
+	gboolean    retval;
+
+	key = panel_gconf_sprintf ("%s/general/%s", config_dir, general_key);
+
+	if (!(entry = gconf_engine_get_entry (engine, key, NULL, FALSE, NULL)))
+		return FALSE;
+
+	retval = gconf_entry_get_value (entry)       != NULL ||
+		 gconf_entry_get_schema_name (entry) != NULL;
+
+	gconf_entry_unref (entry);
+
+	return retval;
+}
+
+static gboolean
+panel_compatibility_detect_needs_migration (void)
+{
+	GConfEngine *engine;
+	gboolean     needs_migration = FALSE;
+
+	if (!(engine = get_homedir_source ()))
+		return FALSE;
+
+	if (!is_general_key_set (engine, PANEL_CONFIG_DIR, "panel_id_list"))
+		goto no_migration;
+
+	if (is_general_key_set (engine, PANEL_CONFIG_DIR, "toplevel_id_list"))
+		goto no_migration;
+
+	needs_migration = TRUE;
+
+ no_migration:
+	gconf_engine_unref (engine);
+
+	return needs_migration;
+}
+
 /* If toplevel_id_list is unset, migrate all the panels in
  * panel_id_list to toplevels
  */
@@ -853,32 +903,17 @@ void
 panel_compatibility_migrate_panel_id_list (GConfClient *client)
 {
 	GHashTable *panel_id_hash;
-	GConfValue *value;
-	GError     *error = NULL;
-	const char *profile;
 	const char *key;
 	GSList     *panel_id_list;
 	GSList     *toplevel_id_list = NULL;
 	GSList     *l;
 
-	profile = panel_profile_get_name ();
-
-	key = panel_gconf_general_key (profile, "toplevel_id_list");
-	value = gconf_client_get (client, key, &error);
-	if (error) {
-		g_warning ("Error reading GConf value from '%s': %s", key, error->message);
-		g_error_free (error);
+	if (!panel_compatibility_detect_needs_migration ())
 		return;
-	}
-
-	if (value) {
-		gconf_value_free (value);
-		return;
-	}
 
 	panel_id_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 
-	key = panel_gconf_general_key (profile, "panel_id_list");
+	key = panel_gconf_general_key ("panel_id_list");
 	panel_id_list = gconf_client_get_list (client, key, GCONF_VALUE_STRING, NULL);
 
 	for (l = panel_id_list; l; l = l->next) {
@@ -897,7 +932,7 @@ panel_compatibility_migrate_panel_id_list (GConfClient *client)
 		g_hash_table_insert (panel_id_hash, l->data, new_id);
 	}
 
-	key = panel_gconf_general_key (profile, "toplevel_id_list");
+	key = panel_gconf_general_key ("toplevel_id_list");
 	gconf_client_set_list (client, key, GCONF_VALUE_STRING, toplevel_id_list, NULL);
 
 	g_slist_free (panel_id_list);
@@ -908,3 +943,125 @@ panel_compatibility_migrate_panel_id_list (GConfClient *client)
 
 	g_hash_table_destroy (panel_id_hash);
 }
+
+static void
+copy_gconf_dir (GConfClient  *client,
+		const char   *src_dir,
+		const char   *dest_dir)
+{
+	GSList *list, *l;
+
+	list = gconf_client_all_entries (client, src_dir, NULL);
+	for (l = list; l; l = l->next) {
+		GConfEntry *entry = l->data;
+		const char *key;
+		char       *tmp;
+
+		tmp = g_path_get_basename (gconf_entry_get_key (entry));
+		key = panel_gconf_sprintf ("%s/%s", dest_dir, tmp);
+		g_free (tmp);
+
+		if (gconf_entry_get_schema_name (entry))
+			gconf_engine_associate_schema (client->engine,
+						       key,
+						       gconf_entry_get_schema_name (entry),
+						       NULL);
+
+		if (entry->value)
+			gconf_client_set (client, key, entry->value, NULL);
+
+		gconf_entry_free (entry);
+	}
+	g_slist_free (list);
+
+	list = gconf_client_all_dirs (client, src_dir, NULL);
+	for (l = list; l; l = l->next) {
+		char *subdir = l->data;
+		char *src_subdir;
+		char *dest_subdir;
+		char *tmp;
+
+		tmp = g_path_get_basename (subdir);
+		src_subdir  = gconf_concat_dir_and_key (src_dir,  tmp);
+		dest_subdir = gconf_concat_dir_and_key (dest_dir, tmp);
+		g_free (tmp);
+
+		copy_gconf_dir (client, src_subdir, dest_subdir);
+
+		g_free (src_subdir);
+		g_free (dest_subdir);
+		g_free (subdir);
+	}
+
+	g_slist_free (list);
+}
+
+void
+panel_compatibility_maybe_copy_old_config (GConfClient *client)
+{
+	GConfEngine *engine;
+	const char  *key;
+
+	key = panel_gconf_general_key ("profiles_migrated");
+	if (gconf_client_get_bool (client, key, NULL))
+		return;
+
+	if (!(engine = get_homedir_source ()))
+		goto no_migration_needed;;
+
+	if (!is_general_key_set (engine, PANEL_OLD_CONFIG_DIR, "panel_id_list")    &&
+	    !is_general_key_set (engine, PANEL_OLD_CONFIG_DIR, "toplevel_id_list") &&
+	    !is_general_key_set (engine, PANEL_OLD_CONFIG_DIR, "applet_id_list")   &&
+	    !is_general_key_set (engine, PANEL_OLD_CONFIG_DIR, "object_id_list"))
+		goto no_migration_needed;
+
+	copy_gconf_dir (client, PANEL_OLD_CONFIG_DIR, PANEL_CONFIG_DIR);
+
+	key = panel_gconf_general_key ("profiles_migrated");
+	gconf_client_set_bool (client, key, TRUE, NULL);
+
+ no_migration_needed:
+	if (engine)
+		gconf_engine_unref (engine);
+}
+
+void
+panel_compatibility_migrate_applications_scheme (GConfClient *client,
+						 const char  *key)
+{
+	char *location;
+
+	location = gconf_client_get_string (client, key, NULL);
+
+	if (!location)
+		return;
+
+	if (!strncmp (location, "applications:", strlen ("applications:")) ||
+	    !strncmp (location, "applications-all-users:", strlen ("applications-all-users:")) ||
+	    !strncmp (location, "all-applications:", strlen ("all-applications:")) ||
+	    !strncmp (location, "preferences:", strlen ("preferences:")) ||
+	    !strncmp (location, "preferences-all-users:", strlen ("preferences-all-users:")) ||
+	    !strncmp (location, "all-preferences:", strlen ("all-preferences:")) ||
+	    !strncmp (location, "system-settings:", strlen ("system-settings:")) ||
+	    !strncmp (location, "server-settings:", strlen ("server-settings:"))) {
+		char *buffer;
+		char *basename;
+		char *new_location;
+
+		basename = g_path_get_basename (location);
+
+		buffer = g_strconcat ("applications/", basename, NULL);
+		g_free (basename);
+
+		new_location = panel_lookup_in_data_dirs (buffer);
+		g_free (buffer);
+
+		if (new_location != NULL) {
+			gconf_client_set_string (client, key,
+						 new_location, NULL);
+			g_free (new_location);
+		}
+	}
+
+	g_free (location);
+}
