@@ -80,6 +80,8 @@ typedef struct {
 	
 	GtkListStore     *program_list_store;
 
+	GHashTable       *dir_hash;
+	GList		 *possible_executables;
 	GList		 *completion_items;
 	GCompletion      *completion;
 	
@@ -138,6 +140,15 @@ panel_run_dialog_destroy (PanelRunDialog *dialog)
 					    dialog->content_notify_id);
 	dialog->content_notify_id = 0;
 
+	if (dialog->dir_hash)
+		g_hash_table_destroy (dialog->dir_hash);
+	dialog->dir_hash = NULL;
+
+	for (l = dialog->possible_executables; l; l = l->next)
+		g_free (l->data);
+	g_list_free (dialog->possible_executables);
+	dialog->possible_executables = NULL;
+	
 	for (l = dialog->completion_items; l; l = l->next)
 		g_free (l->data);
 	g_list_free (dialog->completion_items);
@@ -150,44 +161,29 @@ panel_run_dialog_destroy (PanelRunDialog *dialog)
 	g_free (dialog);
 }
 
-static void
-set_window_icon_from_stock (GtkWindow *window,
-			    const char *stock_id)
-{
-	GdkPixbuf *icon;
-
-	icon = gtk_widget_render_icon (GTK_WIDGET (window),
-				       stock_id,
-				       GTK_ICON_SIZE_DIALOG,
-				       NULL);
-
-	gtk_window_set_icon (window, icon);
-
-	g_object_unref (icon);
-}
-
 static void
 panel_run_dialog_set_icon (PanelRunDialog *dialog,
 			   const char     *icon_path)
 {
 	GdkPixbuf *pixbuf = NULL;
-	char      *icon = NULL;
 
 	g_free (dialog->icon_path);
 	dialog->icon_path = NULL;
 		
-	if (icon_path)
-		icon = gnome_desktop_item_find_icon (panel_icon_theme,
-						     icon_path,
-						     ICON_SIZE /* desired size */,
-						     0 /* flags */);
-						     
-	if (icon)
-		pixbuf = gdk_pixbuf_new_from_file_at_size (icon, ICON_SIZE, ICON_SIZE, NULL);
+	if (icon_path && dialog->icon_path &&
+	    !strcmp (icon_path, dialog->icon_path))
+		return;
 
-                        
+	if (icon_path)
+		pixbuf = panel_load_icon (gtk_icon_theme_get_default (),
+					  icon_path,
+					  ICON_SIZE,
+					  ICON_SIZE,
+					  ICON_SIZE,
+					  NULL);
+	
 	if (pixbuf) {
-		dialog->icon_path = icon;
+		dialog->icon_path = g_strdup (icon_path);
 
 		/* Don't bother scaling the image if it's too small.
 		 * Scaled looks worse than a smaller image.
@@ -196,20 +192,18 @@ panel_run_dialog_set_icon (PanelRunDialog *dialog,
 
 		gtk_window_set_icon (GTK_WINDOW (dialog->run_dialog), pixbuf);
 		
-		/* Don't unref pixbuf here, GTK will do it for us. */
-		gtk_drag_source_set_icon_pixbuf (dialog->run_dialog, pixbuf);		
+		gtk_drag_source_set_icon_pixbuf (dialog->run_dialog, pixbuf);
+		g_object_unref (pixbuf);
 	} else {
-		g_free (icon);
+		gtk_image_set_from_icon_name (GTK_IMAGE (dialog->pixmap),
+					      PANEL_RUN_ICON,
+					      GTK_ICON_SIZE_DIALOG);
 		
-		gtk_image_set_from_stock (GTK_IMAGE (dialog->pixmap), 
-					  PANEL_STOCK_RUN,
-					  GTK_ICON_SIZE_DIALOG);
+		gtk_window_set_icon_name (GTK_WINDOW (dialog->run_dialog),
+					  PANEL_RUN_ICON);
 		
-		set_window_icon_from_stock (GTK_WINDOW (dialog->run_dialog),
-					    PANEL_STOCK_RUN);
-		
-		gtk_drag_source_set_icon_stock (dialog->run_dialog,
-						PANEL_STOCK_RUN);
+		/* FIXME: waiting for bug #116577
+		gtk_drag_source_set_icon_name (dialog->run_dialog, PANEL_RUN_ICON); */
 	}
 }
 
@@ -289,7 +283,7 @@ panel_run_dialog_launch_command (PanelRunDialog *dialog,
 				      &error);
 			
 	if (!result) {
-		panel_error_dialog (screen, "cannot_spawn_command",
+		panel_error_dialog (screen, "cannot_spawn_command", TRUE,
 				    _("Cannot launch command '%s'"),
 				    "%s",
 				    escaped,
@@ -315,7 +309,7 @@ panel_run_dialog_show_url (PanelRunDialog *dialog,
 
 	gnome_url_show_on_screen (url, screen, &error);
 	if (error) {
-		panel_error_dialog (screen, "cannot_show_url",
+		panel_error_dialog (screen, "cannot_show_url", TRUE,
 				    _("Cannot display location '%s'"),
 				    "%s",
 				    escaped,
@@ -368,7 +362,7 @@ panel_run_dialog_execute (PanelRunDialog *dialog)
 
 	if (!disk || error) {
 		panel_error_dialog (screen, "cannot_convert_command_from_utf8",
-				    _("Cannot convert '%s' from UTF-8"),
+				    TRUE, _("Cannot convert '%s' from UTF-8"),
 				    "%s",
 				    command,
 				    error->message);
@@ -377,21 +371,21 @@ panel_run_dialog_execute (PanelRunDialog *dialog)
 		return;
 	}
 
-	url = gnome_vfs_make_uri_from_input_with_dirs (disk,
+	url = gnome_vfs_make_uri_from_input_with_dirs (command,
 						       GNOME_VFS_MAKE_URI_DIR_HOMEDIR);
 	escaped = g_markup_escape_text (url, -1);
 	scheme = gnome_vfs_get_uri_scheme (url);
 	result = FALSE;
 	
-	if (!strcasecmp (scheme, "http") ||
-	    !strcasecmp (scheme, "file"))
+	if (!g_ascii_strcasecmp (scheme, "http") ||
+	    !g_ascii_strcasecmp (scheme, "file"))
 		/* If this returns an http or file url, the url might refer to a
 		 * command that is somewhere in the path or an executable file.
 		 * So try executing it before displaying it. We execute the 
 		 * command in the user's shell so that it can do all the parameter
 		 * expansion and other magic for us.
 		 */
-		result = panel_run_dialog_launch_command (dialog, command, escaped);
+		result = panel_run_dialog_launch_command (dialog, disk, escaped);
 	
 	if (!result)
 		result = panel_run_dialog_show_url (dialog, url, escaped);
@@ -584,19 +578,6 @@ panel_run_dialog_find_command_icon_idle (PanelRunDialog *dialog)
 				    COLUMN_PATH,      &path,
 				    -1);
 
-		if (!exec) {
-			GnomeDesktopItem *item;
-
-			if ((item = gnome_desktop_item_new_from_file (path, 0, NULL))) {
-				exec = g_strdup (gnome_desktop_item_get_string (item, GNOME_DESKTOP_ITEM_EXEC));
-				gnome_desktop_item_unref (item);
-			}
-
-			gtk_list_store_set (GTK_LIST_STORE (model), &iter,
-					    COLUMN_EXEC, exec ? exec : "",
-					    -1);
-		}
-
         	if (exec && icon) {
 			gboolean fuzzy = FALSE;
 
@@ -679,7 +660,7 @@ panel_run_dialog_add_icon_idle (PanelRunDialog *dialog)
 			icon_height = PANEL_DEFAULT_MENU_ICON_SIZE;
 		}
 
-		pixbuf = panel_make_menu_icon (file, NULL, icon_height, &long_operation);
+		pixbuf = panel_make_menu_icon (gtk_icon_theme_get_default (), file, NULL, icon_height, &long_operation);
 		if (pixbuf) {
 			gtk_list_store_set (dialog->program_list_store, &iter, COLUMN_ICON, pixbuf, -1);
 			g_object_unref (pixbuf);
@@ -693,32 +674,74 @@ panel_run_dialog_add_icon_idle (PanelRunDialog *dialog)
 }
 
 static int
-compare_applications (MenuTreeEntry *a,
-		      MenuTreeEntry *b)
+compare_applications (GMenuTreeEntry *a,
+		      GMenuTreeEntry *b)
 {
-	return g_utf8_collate (menu_tree_entry_get_name (a),
-			       menu_tree_entry_get_name (b));
+	return g_utf8_collate (gmenu_tree_entry_get_name (a),
+			       gmenu_tree_entry_get_name (b));
+}
+
+static GSList *get_all_applications_from_dir (GMenuTreeDirectory *directory,
+					      GSList            *list);
+
+static GSList *
+get_all_applications_from_alias (GMenuTreeAlias *alias,
+				 GSList         *list)
+{
+	GMenuTreeItem *aliased_item;
+
+	aliased_item = gmenu_tree_alias_get_item (alias);
+
+	switch (gmenu_tree_item_get_type (aliased_item)) {
+	case GMENU_TREE_ITEM_ENTRY:
+		list = g_slist_append (list, gmenu_tree_item_ref (aliased_item));
+		break;
+
+	case GMENU_TREE_ITEM_DIRECTORY:
+		list = get_all_applications_from_dir (GMENU_TREE_DIRECTORY (aliased_item),
+						      list);
+		break;
+
+	default:
+		break;
+	}
+
+	gmenu_tree_item_unref (aliased_item);
+
+	return list;
 }
 
 static GSList *
-get_all_applications_from_dir (MenuTreeDirectory *directory,
-			       GSList            *list)
+get_all_applications_from_dir (GMenuTreeDirectory *directory,
+			       GSList             *list)
 {
-	GSList *subdirs;
+	GSList *items;
 	GSList *l;
 
-	list = g_slist_concat (list,
-			       menu_tree_directory_get_entries (directory));
+	items = gmenu_tree_directory_get_contents (directory);
 
-	subdirs = menu_tree_directory_get_subdirs (directory);
-	for (l = subdirs; l; l = l->next) {
-		MenuTreeDirectory *subdir = l->data;
+	for (l = items; l; l = l->next) {
+		switch (gmenu_tree_item_get_type (l->data)) {
+		case GMENU_TREE_ITEM_ENTRY:
+			list = g_slist_append (list, gmenu_tree_item_ref (l->data));
+			break;
 
-		list = get_all_applications_from_dir (subdir, list);
+		case GMENU_TREE_ITEM_DIRECTORY:
+			list = get_all_applications_from_dir (l->data, list);
+			break;
 
-		menu_tree_directory_unref (subdir);
+		case GMENU_TREE_ITEM_ALIAS:
+			list = get_all_applications_from_alias (l->data, list);
+			break;
+
+		default:
+			break;
+		}
+
+		gmenu_tree_item_unref (l->data);
 	}
-	g_slist_free (subdirs);
+
+	g_slist_free (items);
 
 	return list;
 }
@@ -726,18 +749,18 @@ get_all_applications_from_dir (MenuTreeDirectory *directory,
 static GSList *
 get_all_applications (void)
 {
-	MenuTree          *tree;
-	MenuTreeDirectory *root;
-	GSList            *retval;
+	GMenuTree          *tree;
+	GMenuTreeDirectory *root;
+	GSList             *retval;
 
-	tree = menu_tree_lookup ("applications.menu");
+	tree = gmenu_tree_lookup ("applications.menu", GMENU_TREE_FLAGS_NONE);
 
-	root = menu_tree_get_root_directory (tree);
+	root = gmenu_tree_get_root_directory (tree);
 
 	retval = get_all_applications_from_dir (root, NULL);
 
-	menu_tree_directory_unref (root);
-	menu_tree_unref (tree);
+	gmenu_tree_item_unref (root);
+	gmenu_tree_unref (tree);
 
 	retval = g_slist_sort (retval,
 			       (GCompareFunc) compare_applications);
@@ -769,38 +792,39 @@ panel_run_dialog_add_items_idle (PanelRunDialog *dialog)
 	/* Strip duplicates */
 	prev_name = NULL;
 	for (l = all_applications; l; l = next) {
-		MenuTreeEntry *entry = l->data;
+		GMenuTreeEntry *entry = l->data;
 
 		next = l->next;
 
-		if (prev_name && strcmp (menu_tree_entry_get_name (entry), prev_name) == 0) {
-			menu_tree_entry_unref (entry);
+		if (prev_name && strcmp (gmenu_tree_entry_get_name (entry), prev_name) == 0) {
+			gmenu_tree_item_unref (entry);
 
 			all_applications = g_slist_delete_link (all_applications, l);
 		} else {
-			prev_name = menu_tree_entry_get_name (entry);
+			prev_name = gmenu_tree_entry_get_name (entry);
 		}
 	}
 
 	for (l = all_applications; l; l = l->next) {
-		MenuTreeEntry *entry = l->data;
+		GMenuTreeEntry *entry = l->data;
 		GtkTreeIter    iter;
 		GtkTreePath   *path;
 
 		gtk_list_store_append (dialog->program_list_store, &iter);
 		gtk_list_store_set (dialog->program_list_store, &iter,
 				    COLUMN_ICON,      NULL,
-				    COLUMN_ICON_FILE, menu_tree_entry_get_icon (entry),
-				    COLUMN_NAME,      menu_tree_entry_get_name (entry),
-				    COLUMN_COMMENT,   menu_tree_entry_get_comment (entry),
-				    COLUMN_PATH,      menu_tree_entry_get_desktop_file_path (entry),
+				    COLUMN_ICON_FILE, gmenu_tree_entry_get_icon (entry),
+				    COLUMN_NAME,      gmenu_tree_entry_get_name (entry),
+				    COLUMN_COMMENT,   gmenu_tree_entry_get_comment (entry),
+				    COLUMN_EXEC,      gmenu_tree_entry_get_exec (entry),
+				    COLUMN_PATH,      gmenu_tree_entry_get_desktop_file_path (entry),
 				    -1);
 
 		path = gtk_tree_model_get_path (GTK_TREE_MODEL (dialog->program_list_store), &iter);
 		if (path != NULL)
 			dialog->add_icon_paths = g_slist_prepend (dialog->add_icon_paths, path);
 
-		menu_tree_entry_unref (entry);
+		gmenu_tree_item_unref (entry);
 	}
 	g_slist_free (all_applications);
 
@@ -1116,39 +1140,48 @@ panel_run_dialog_setup_file_button (PanelRunDialog *dialog,
 }
 
 static GList *
-fill_files_from (GList      *list,
-		 const char *dirname,
-		 const char *prefix,
+fill_files_from (const char *dirname,
+		 const char *dirprefix,
+		 char        prefix,
 		 GList      *existing_items)
 {
-	struct dirent *dent;
+	GList         *list;
 	DIR           *dir;
+	struct dirent *dent;
 	
+	list = NULL;
 	dir = opendir (dirname);
 	
 	if (!dir)
 		return list;
 	
 	while ((dent = readdir (dir))) {
-		char          *file;
-		char          *item;
-		const char    *suffix;
+		char       *file;
+		char       *item;
+		const char *suffix;
 		
+		if (!dent->d_name ||
+		    dent->d_name [0] != prefix)
+			continue;
+
 		file = g_build_filename (dirname, dent->d_name, NULL);
 		
 		suffix = NULL;
-		if (g_file_test (file, G_FILE_TEST_IS_DIR))
+		if (
+#ifdef HAVE_STRUCT_DIRENT_D_TYPE
+		/* don't use g_file_test at first so we don't stat() */
+		    dent->d_type == DT_DIR ||
+		    (dent->d_type == DT_LNK &&
+		     g_file_test (file, G_FILE_TEST_IS_DIR))
+#else
+		    g_file_test (file, G_FILE_TEST_IS_DIR)
+#endif
+		   )
 			suffix = "/";
 		
 		g_free (file);
 		
-		item = g_strconcat (prefix, dent->d_name, suffix, NULL);
-		
-		if (g_list_find_custom (existing_items, item, (GCompareFunc) strcmp)) {
-			g_free (item);
-			closedir (dir);
-			return list;
-		}
+		item = g_build_filename (dirprefix, dent->d_name, suffix, NULL);
 		
 		list = g_list_prepend (list, item);
 	}
@@ -1159,100 +1192,152 @@ fill_files_from (GList      *list,
 }	
 
 static GList *
-fill_executables_from (GList      *list,
-		       const char *dirname)
+fill_possible_executables (void)
 {
-	struct dirent *dent;
-	DIR           *dir;
-	char          *file;
-	struct stat    file_stat;
-
-	dir = opendir (dirname);
-
-	if (!dir)
-		return list;
-
-	while ((dent = readdir (dir))) {
-		file = g_build_filename (dirname, dent->d_name, NULL);
-
-		if (!access (file, X_OK) &&
-		    !stat (file, &file_stat) && !S_ISDIR (file_stat.st_mode))
-			list = g_list_prepend (list, g_strdup (dent->d_name));
-		
-		g_free (file);
-	}
-
-	closedir (dir);
+	GList         *list;
+	const char    *path;
+	char         **pathv;
+	int            i;
 	
-	return list;
-}
-
-static GList *
-fill_executables (GList *list)
-{
-	const char  *path;
-	char       **pathv;
-	int          i;
-
+	list = NULL;
 	path = g_getenv ("PATH");
 
 	if (!path || !path [0])
 		return list;
 
 	pathv = g_strsplit (path, ":", 0);
+	
+	for (i = 0; pathv [i]; i++) {
+		const char *file;
+		char       *filename;
+		GDir       *dir;
 
-	for (i = 0; pathv [i]; i++)
-		list = fill_executables_from (list, pathv [i]);
+		dir = g_dir_open (pathv [i], 0, NULL);
+
+		if (!dir)
+			continue;
+
+		while ((file = g_dir_read_name (dir))) {
+			filename = g_build_filename (pathv [i], file, NULL);
+			list = g_list_prepend (list, filename);
+		}
+
+		g_dir_close (dir);
+	}
 	
 	g_strfreev (pathv);
 	
 	return list;
 }
 
-static void
-panel_run_dialog_ensure_completion (PanelRunDialog *dialog)
+static GList *
+fill_executables (GList *possible_executables,
+		  GList *existing_items,
+		  char   prefix)
 {
-	if (!dialog->completion) {
-		dialog->completion = g_completion_new (NULL);
-		
-		dialog->completion_items = fill_executables (NULL);
-		dialog->completion_items = fill_files_from  (dialog->completion_items, "/", "/", NULL);
-		dialog->completion_items = fill_files_from  (dialog->completion_items, g_get_home_dir(), "", NULL);
-		
-		g_completion_add_items (dialog->completion,
-					dialog->completion_items);
+	GList *list;
+	GList *l;
+	
+	list = NULL;	
+	
+	for (l = possible_executables; l; l = l->next) {
+		const char *filename;
+		char       *basename;
+			
+		filename = l->data;
+		basename = g_path_get_basename (filename);
+			
+		if (basename [0] == prefix &&
+		    g_file_test (filename, G_FILE_TEST_IS_REGULAR) &&
+		    g_file_test (filename, G_FILE_TEST_IS_EXECUTABLE)) {
+			    
+			if (g_list_find_custom (existing_items, basename,
+						(GCompareFunc) strcmp)) {
+				g_free (basename);
+				return NULL;
+			}
+
+			list = g_list_prepend (list, basename);
+		 } else {
+			g_free (basename);
+		 }
 	}
+	
+	return list;
 }
 
 static void
 panel_run_dialog_update_completion (PanelRunDialog *dialog,
 				    const char     *text)
 {
-	g_assert (dialog->completion != NULL);
-	
-	if (text [strlen (text) - 1] == '/') {
-		const char *dirname;
-		char       *freeme;
-		GList      *list;
-		
-		dirname = text;
-		freeme = NULL;
-		
-		if (dirname [0] != '/')
-			dirname = freeme = g_build_filename(g_get_home_dir(), text, NULL);
-		
-		list = fill_files_from (NULL, dirname, text, dialog->completion_items);
-		
-		if (freeme)
-			g_free (freeme);
-		
-		if (list == NULL)
-			return;
-		
-		g_completion_add_items (dialog->completion, list);
-		
-		dialog->completion_items = g_list_concat (dialog->completion_items, list);
+	GList *list;
+	GList *executables;
+	char   prefix;
+	char  *buf;
+	char  *dirname;
+	char  *dirprefix;
+	char  *key;
+
+	g_assert (text != NULL && *text != '\0' && !g_ascii_isspace (*text));
+
+	list = NULL;
+	executables = NULL;
+
+	if (!dialog->completion) {
+		dialog->completion = g_completion_new (NULL);
+		dialog->possible_executables = fill_possible_executables ();
+		dialog->dir_hash = g_hash_table_new_full (g_str_hash,
+							  g_str_equal,
+							  g_free, NULL);
 	}
+	
+	buf = g_path_get_basename (text);
+	prefix = buf[0];
+	g_free (buf);
+	if (prefix == '/' || prefix == '.')
+		return;
+
+	if (text [0] == '/') {
+		/* complete against absolute path */
+		dirname = g_path_get_dirname (text);
+		dirprefix = g_strdup (dirname);
+	} else {
+		/* complete against relative path and executable name */
+		if (!strchr (text, '/')) {
+			executables = fill_executables (dialog->possible_executables,
+							dialog->completion_items,
+							text [0]);
+			dirprefix = g_strdup ("");
+		} else {
+			dirprefix = g_path_get_dirname (text);
+		}
+
+		dirname = g_build_filename (g_get_home_dir (), dirprefix, NULL);
+	}
+
+	key = g_strdup_printf ("%s%c%c", dirprefix, G_DIR_SEPARATOR, prefix);
+
+	if (!g_hash_table_lookup (dialog->dir_hash, key)) {
+		g_hash_table_insert (dialog->dir_hash, key, dialog);
+
+		list = fill_files_from (dirname, dirprefix, prefix,
+					dialog->completion_items);
+	} else {
+		g_free (key);
+	}
+
+	list = g_list_concat (list, executables);
+
+	g_free (dirname);
+	g_free (dirprefix);
+
+	if (list == NULL)
+		return;
+		
+	g_completion_add_items (dialog->completion, list);
+		
+	dialog->completion_items = g_list_concat (dialog->completion_items,
+						  list);	
 }
 
 static gboolean
@@ -1261,17 +1346,22 @@ entry_event (GtkEditable    *entry,
 	     PanelRunDialog *dialog)
 {
 	char *prefix;
+	char *nospace_prefix;
 	char *nprefix;
 	char *temp;
 	int   pos, tmp;
 
+	if (event->type != GDK_KEY_PRESS)
+		return FALSE;
+
 	/* if user typed something we're not using the list anymore */
-	if (event->type == GDK_KEY_PRESS)
-		dialog->use_program_list = FALSE;
+	dialog->use_program_list = FALSE;
+
+	if (!panel_profile_get_enable_autocompletion ())
+		return FALSE;
 
 	/* tab completion */
-	if (event->type == GDK_KEY_PRESS &&
-	    event->keyval == GDK_Tab) {
+	if (event->keyval == GDK_Tab) {
 		gtk_editable_get_selection_bounds (entry, &pos, &tmp);
 
 		if (dialog->completion_started &&
@@ -1280,21 +1370,16 @@ entry_event (GtkEditable    *entry,
 		    tmp == strlen (gtk_entry_get_text (GTK_ENTRY (entry)))) {
 	    		gtk_editable_select_region (entry, 0, 0);		
 			gtk_editable_set_position (entry, -1);
-
-			panel_run_dialog_update_completion (dialog,
-			    				    gtk_entry_get_text (GTK_ENTRY (entry)));
 			
 			return TRUE;
 		}
-	} else if (event->type == GDK_KEY_PRESS &&
-		   event->length > 0) {
-		panel_run_dialog_ensure_completion (dialog);
-
+	} else if (event->length > 0) {
+			   
 		gtk_editable_get_selection_bounds (entry, &pos, &tmp);
 
 		if (dialog->completion_started &&
 		    pos != tmp &&
-		    pos != 1 &&
+		    pos != 0 &&
 		    tmp == strlen (gtk_entry_get_text (GTK_ENTRY (entry)))) {
 			temp = gtk_editable_get_chars (entry, 0, pos);
 			prefix = g_strconcat (temp, event->string, NULL);
@@ -1307,29 +1392,42 @@ entry_event (GtkEditable    *entry,
 			return FALSE;
 		}
 		
+		nospace_prefix = prefix;
+		while (*nospace_prefix != '\0' &&
+		       g_ascii_isspace (*nospace_prefix))
+			nospace_prefix++;
+		if (*nospace_prefix == '\0')
+			return FALSE;
+
+		panel_run_dialog_update_completion (dialog, nospace_prefix);
+		
+		if (!dialog->completion) {
+			g_free (prefix);
+			return FALSE;
+		}
+		
 		pos = strlen (prefix);
 		nprefix = NULL;
 
-		g_completion_complete (dialog->completion, prefix, &nprefix);
+		g_completion_complete (dialog->completion, nospace_prefix,
+				       &nprefix);
 
 		if (nprefix) {
-		    	gtk_entry_set_text (GTK_ENTRY (entry), nprefix);
+			temp = g_strndup (prefix, nospace_prefix - prefix);
+			g_free (prefix);
+
+			prefix = g_strconcat (temp, nprefix, NULL);
+		    	gtk_entry_set_text (GTK_ENTRY (entry), prefix);
  			gtk_editable_set_position (entry, pos);
 			gtk_editable_select_region (entry, pos, -1);
 			
 			dialog->completion_started = TRUE;
 
-			if (strcmp(event->string, "/") == 0)
-				panel_run_dialog_update_completion (dialog, nprefix);
-			
 			g_free (nprefix);
 			g_free (prefix);
 			
 			return TRUE;
 		}
-
-		if (strcmp(event->string, "/") == 0)
-			panel_run_dialog_update_completion (dialog, prefix);
 		
 		g_free (prefix);
 	}
@@ -1341,29 +1439,34 @@ static void
 entry_changed (GtkEntry       *entry,
 	       PanelRunDialog *dialog)
 {
-	static GtkTargetEntry  drag_types[] = { { "text/uri-list", 0, 0 } };
-	const char            *text;
-	char                  *msg;
+	const char *text;
+	const char *start;
+	char       *msg;
 
 	text = gtk_entry_get_text (entry);
 
+	start = text;
+	while (*start != '\0' && g_ascii_isspace (*start))
+		start++;
+
 	/* desensitize run button if no text entered */
-	if (!text || !text [0]) {
+	if (!start || !start [0]) {
 		gtk_widget_set_sensitive (dialog->run_button, FALSE);
 		gtk_drag_source_unset (dialog->run_dialog);
 	} else {
 		gtk_widget_set_sensitive (dialog->run_button, TRUE);
 		gtk_drag_source_set (dialog->run_dialog,
 				     GDK_BUTTON1_MASK,
-				     drag_types,
-				     G_N_ELEMENTS (drag_types),
+				     NULL, 0,
 				     GDK_ACTION_COPY);
+		gtk_drag_source_add_uri_targets (dialog->run_dialog);
 	}
 	
 	/* update description label */
-        if (text && text [0]) {
+        if (start && start [0]) {
 		if (!dialog->use_program_list) {
-			msg = g_strdup_printf (_("Will run command: '%s'"), text);
+			msg = g_strdup_printf (_("Will run command: '%s'"),
+					       start);
 			gtk_label_set_text (GTK_LABEL (dialog->program_label), msg);
 			g_free (msg);
 		}
@@ -1378,7 +1481,7 @@ entry_changed (GtkEntry       *entry,
 	}
 
 	/* look up icon for the command */
-	if (text && text [0] &&
+	if (start && start [0] &&
 	    panel_profile_get_enable_program_list () &&
 	    !dialog->use_program_list &&
 	    !dialog->find_command_icon_idle_id)
@@ -1398,13 +1501,23 @@ entry_drag_data_received (GtkEditable      *entry,
 			  guint32           time,
 			  PanelRunDialog   *dialog)
 {
+	char  *uri_list;
 	char **uris;
 	char  *file;
 	int    i;
 
-	uris = g_strsplit (selection_data->data, "\r\n", -1);
+        if (selection_data->format != 8 || selection_data->length == 0) {
+        	g_warning (_("URI list dropped on run dialog had wrong format (%d) or length (%d)\n"),
+			   selection_data->format,
+			   selection_data->length);
+		return;
+        }
+
+	uri_list = g_strndup (selection_data->data, selection_data->length);
+	uris = g_strsplit (uri_list, "\r\n", -1);
 
 	if (!uris) {
+		g_free (uri_list);
 		gtk_drag_finish (context, FALSE, FALSE, time);
 		return;
 	}
@@ -1424,6 +1537,7 @@ entry_drag_data_received (GtkEditable      *entry,
 	}
 
 	g_strfreev (uris);
+	g_free (uri_list);
 	gtk_drag_finish (context, TRUE, FALSE, time);
 }
 
@@ -1431,7 +1545,6 @@ static void
 panel_run_dialog_setup_entry (PanelRunDialog *dialog,
 			      GladeXML       *gui)
 {
-	static GtkTargetEntry  drop_types[] = { { "text/uri-list", 0, 0 } };
 	GdkScreen             *screen;
 	int                    width_request;
 	
@@ -1446,7 +1559,7 @@ panel_run_dialog_setup_entry (PanelRunDialog *dialog,
 		      "width_request", width_request,
 		      NULL);
 
-        g_signal_connect (dialog->gtk_entry, "event",
+        g_signal_connect (dialog->gtk_entry, "key-press-event",
 			  G_CALLBACK (entry_event), dialog);
 			  
         g_signal_connect (dialog->gtk_entry, "changed",
@@ -1455,10 +1568,10 @@ panel_run_dialog_setup_entry (PanelRunDialog *dialog,
 	gtk_drag_dest_unset (dialog->gtk_entry);
 	
 	gtk_drag_dest_set (dialog->gtk_entry,
-			   GTK_DEST_DEFAULT_ALL,
-			   drop_types,
-			   G_N_ELEMENTS (drop_types),
+			   GTK_DEST_DEFAULT_MOTION|GTK_DEST_DEFAULT_HIGHLIGHT,
+			   NULL, 0,
 			   GDK_ACTION_COPY);
+	gtk_drag_dest_add_uri_targets (dialog->gtk_entry);
 
 	g_signal_connect (dialog->gtk_entry, "drag_data_received",
 			  G_CALLBACK (entry_drag_data_received), dialog);
@@ -1491,10 +1604,10 @@ pixmap_drag_data_get (GtkWidget          *run_dialog,
 						       GNOME_VFS_MAKE_URI_DIR_HOMEDIR);
 	scheme = gnome_vfs_get_uri_scheme (uri);
 	
-	if (!strcasecmp (scheme, "http"))
+	if (!g_ascii_strcasecmp (scheme, "http"))
 		exec = command_is_executable (text);
 		
-	else if (!strcasecmp (scheme, "file"))
+	else if (!g_ascii_strcasecmp (scheme, "file"))
 		exec = command_is_executable (uri);
 		
 	if (exec) {
@@ -1543,10 +1656,6 @@ panel_run_dialog_setup_pixmap (PanelRunDialog *dialog,
 	g_signal_connect (dialog->run_dialog, "drag_data_get",
 			  G_CALLBACK (pixmap_drag_data_get),
 			  dialog);
-
-	gtk_image_set_from_stock (GTK_IMAGE (dialog->pixmap), 
-				  PANEL_STOCK_RUN,
-				  GTK_ICON_SIZE_DIALOG);
 }
 
 static PanelRunDialog *
@@ -1575,6 +1684,8 @@ panel_run_dialog_new (GdkScreen *screen,
 	panel_run_dialog_setup_program_list  (dialog, gui);
 	panel_run_dialog_setup_list_expander (dialog, gui);
 
+	panel_run_dialog_set_icon            (dialog, NULL);
+
 	panel_run_dialog_update_content (dialog, panel_profile_get_show_program_list ());
 
 	gtk_widget_set_sensitive (dialog->run_button, FALSE);
@@ -1584,11 +1695,10 @@ panel_run_dialog_new (GdkScreen *screen,
 
 	gtk_window_set_screen (GTK_WINDOW (dialog->run_dialog), screen);
 
-	set_window_icon_from_stock (GTK_WINDOW (dialog->run_dialog), PANEL_STOCK_RUN);
-
 	gtk_widget_grab_focus (dialog->gtk_entry);
 	gtk_widget_realize (dialog->run_dialog);
-	panel_gdk_x11_window_set_user_time (dialog->run_dialog->window, activate_time);
+	gdk_x11_window_set_user_time (dialog->run_dialog->window,
+				      activate_time);
 	gtk_widget_show (dialog->run_dialog);
 	
 	return dialog;
@@ -1612,7 +1722,8 @@ panel_run_dialog_present (GdkScreen *screen,
 
 	if (static_dialog) {
 		gtk_window_set_screen (GTK_WINDOW (static_dialog->run_dialog), screen);
-		panel_gdk_x11_window_set_user_time (static_dialog->run_dialog->window, activate_time);
+		gdk_x11_window_set_user_time (static_dialog->run_dialog->window,
+					      activate_time);
 		gtk_window_present (GTK_WINDOW (static_dialog->run_dialog));
 		gtk_widget_grab_focus (static_dialog->gtk_entry);
 		return;
